@@ -259,8 +259,7 @@ class DataTypes:
 
     @staticmethod
     def str_list(schema: dict, ptyp2: dict or None = None) -> dict:
-        return DataTypes.list({
-            'schema': schema,
+        return DataTypes.type(DataTypes.list(schema, {
             'cast': {
                 'type': 'str',
                 'join': ','
@@ -269,7 +268,7 @@ class DataTypes:
                 lang.EN: "{name} must be a string comma-separated list",
                 lang.ES: "{name} debe ser una lista de elementos separados por comas",
             }
-        }, ptyp2)
+        }), ptyp2)
 
     @staticmethod
     def dict(ptyp2: dict or None = None) -> dict:
