@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 # pylint: disable=missing-docstring, C0301
 import re
+from datetime import date, time, datetime
 
 from dataskema import util
 from dataskema import lang
@@ -17,9 +18,12 @@ KEYWORD_DESCRIPTION = 'description'
 KEYWORD_SEPARATOR = 'separator'
 KEYWORD_MANDATORY = 'mandatory'
 KEYWORD_DEFAULT = 'default'
+
 KEYWORD_TYPE = 'type'
 KEYWORD_TYPE_LIST = 'list'
 KEYWORD_TYPE_DICT = 'dict'
+KEYWORD_IEMPTY = 'iempty'
+KEYWORD_IDUPLICATED = 'iduplicated'
 
 KEYWORD_TYPE_STR = 'str'
 KEYWORD_MIN_SIZE = 'min-size'
@@ -28,6 +32,8 @@ KEYWORD_REGEXP = 'regexp'
 KEYWORD_ICASE = 'icase'
 KEYWORD_WHITE_LIST = 'white-list'
 KEYWORD_MAX_LINES = 'max-lines'
+KEYWORD_FORMAT = 'format'
+KEYWORD_JOIN = 'join'
 
 KEYWORD_TYPE_INT = 'int'
 KEYWORD_TYPE_FLOAT = 'float'
@@ -47,6 +53,15 @@ KEYWORD_TO_FUNC_UPPER = 'upper'
 KEYWORD_TO_FUNC_NO_TRIM = 'no-trim'
 KEYWORD_TO_FUNC_TRIM = 'trim'
 
+KEYWORD_CAST = 'cast'
+
+KEYWORD_TYPE_DATE = 'date'
+KEYWORD_TYPE_TIME = 'time'
+KEYWORD_TYPE_DATETIME = 'datetime'
+DEFAULT_DATE_FORMAT = '%Y-%m-%d'
+DEFAULT_TIME_FORMAT = '%H:%M:%S'
+DEFAULT_DATETIME_FORMAT = DEFAULT_DATE_FORMAT + ' ' + DEFAULT_TIME_FORMAT
+
 ERROR_INVALID_TYPE = "'{keyword}' has an invalid value '{itype}'"
 ERROR_INVALID_DEFAULT_TYPE = "Invalid default value type '{def_itype}'. Expected type '{itype}"
 ERROR_UNEXPECTED_TYPE = "Type '{expected_type}' expected. Found '{utype}'"
@@ -105,13 +120,32 @@ class SchemaValidator:
         if itype is not None:
             try:
                 if itype == KEYWORD_TYPE_STR:
-                    value = _get_type_str(label, value)
+                    keyword2 = KEYWORD_FORMAT
+                    sformat = self._get_str(keyword2)
+                    keyword2 = KEYWORD_JOIN
+                    join = self._get_str(keyword2)
+                    value = _get_type_str(label, value, sformat, join)
                 elif itype == KEYWORD_TYPE_INT:
                     value = _get_type_int(label, value)
                 elif itype == KEYWORD_TYPE_FLOAT:
                     value = _get_type_float(label, value)
                 elif itype == KEYWORD_TYPE_BOOL:
                     value = _get_type_bool(label, value)
+                elif itype == KEYWORD_TYPE_DATE:
+                    keyword2 = KEYWORD_FORMAT
+                    date_format = self._get_str(keyword2)
+                    date_format = date_format if date_format is not None else DEFAULT_DATE_FORMAT
+                    value = _get_type_date(label, value, date_format)
+                elif itype == KEYWORD_TYPE_TIME:
+                    keyword2 = KEYWORD_FORMAT
+                    time_format = self._get_str(keyword2)
+                    time_format = time_format if time_format is not None else DEFAULT_TIME_FORMAT
+                    value = _get_type_time(label, value, time_format)
+                elif itype == KEYWORD_TYPE_DATETIME:
+                    keyword2 = KEYWORD_FORMAT
+                    datetime_format = self._get_str(keyword2)
+                    datetime_format = datetime_format if datetime_format is not None else DEFAULT_DATETIME_FORMAT
+                    value = _get_type_datetime(label, value, datetime_format)
                 elif itype == KEYWORD_TYPE_LIST:
                     keyword2 = KEYWORD_SEPARATOR
                     separator = self._get_str(keyword2)
@@ -141,21 +175,39 @@ class SchemaValidator:
         if mandatory:
             _check_mandatory(label, value)
 
+        # -- aplica las funciones de transformación
+        keyword = KEYWORD_TO
+        tos = self._get_str(keyword)
+        if tos is not None:
+            tos = tos.split(',')
+            for to in tos:
+                to = util.trim(to)
+                if isinstance(value, str):
+                    value = _process_to(keyword, value, to)
+                if isinstance(value, list):
+                    for (pos, item) in enumerate(value):
+                        item = _process_to(keyword, item, to)
+                        value[pos] = item
+                if isinstance(value, dict):
+                    for mkey in value.keys():
+                        item = value.get(mkey)
+                        item = _process_to(keyword, item, to)
+                        value[mkey] = item
+        else:
+            if isinstance(value, str):
+                value = util.trim(value)
+            if isinstance(value, list):
+                for (pos, item) in enumerate(value):
+                    if isinstance(item, str):
+                        value[pos] = util.trim(item)
+            if isinstance(value, dict):
+                for mkey in value.keys():
+                    item = value.get(mkey)
+                    if isinstance(item, str):
+                        value[mkey] = util.trim(item)
+
         # -- comprueba el tipo 'str'
         if isinstance(value, str):
-
-            # -- aplica las funciones de transformación
-            keyword = KEYWORD_TO
-            tos = self._get_str(keyword)
-            if tos is not None:
-                tos = tos.split(',')
-                for to in tos:
-                    to = util.trim(to)
-                    if not util.is_empty(to):
-                        value = _process_to(keyword, value, to)
-            else:
-                value = util.trim(value)
-
             keyword = KEYWORD_MIN_SIZE
             min_size = self._get_int(keyword)
             _check_minsize(label, value, min_size)
@@ -193,10 +245,35 @@ class SchemaValidator:
 
         # -- comprueba el tipo 'list'
         if isinstance(value, list):
+            keyword2 = KEYWORD_IEMPTY
+            iempty = self._get_bool(keyword2)
+            if iempty:
+                outlist = []
+                for item in value:
+                    if item is None or item == '':
+                        continue
+                    outlist.append(item)
+                value = outlist
+            keyword2 = KEYWORD_IDUPLICATED
+            iduplicated = self._get_bool(keyword2)
+            if iduplicated:
+                outlist = []
+                for item in value:
+                    if item in outlist:
+                        continue
+                    outlist.append(item)
+                value = outlist
             keyword = KEYWORD_SCHEMA
             item_schema = self._get_dict(keyword)
             _check_item_list_schema(label, value, item_schema)
 
+        # -- aplica las funciones de casting
+        keyword = KEYWORD_CAST
+        cast = self._get_dict(keyword)
+        if cast is not None:
+            validator = SchemaValidator(cast)
+            value = validator.validate(name, value)
+
         return value
 
     def _get_obj(self, keyword: str):
@@ -247,7 +324,9 @@ class SchemaValidator:
             message = message_map
         else:
             raise fail
-        raise SchemaValidationFailure(message, {KEYWORD_NAME: label})
+        raise SchemaValidationFailure(message, {
+            KEYWORD_NAME: label,
+        })
 
 
 def _schema_error_unexpected_type(keyword: str, found_value: any, expected_type: str):
@@ -260,13 +339,54 @@ def _val_error_unexpected_type(name: str, found_value: any, expected_type: str):
     raise SchemaValidationFailure(lang.VAL_ERROR_PARAM_HAS_INVALID_TYPE, {KEYWORD_NAME: name, 'utype': utype, 'expected_type': expected_type})
 
 
-def _get_type_str(name: str, value: any) -> str or None:
+def  _format_datetime_format(dformat: str) -> str:
+    dformat = dformat.replace('%Y', 'YYYY')
+    dformat = dformat.replace('%m', 'MM')
+    dformat = dformat.replace('%d', 'DD')
+    dformat = dformat.replace('%H', 'HH')
+    dformat = dformat.replace('%M', 'MM')
+    dformat = dformat.replace('%S', 'SS')
+    return dformat
+
+
+def _val_error_invalid_date_format(name: str, value: any, date_format: str):
+    date_format = _format_datetime_format(date_format)
+    raise SchemaValidationFailure(lang.VAL_ERROR_PARAM_DATE_FORMAT, {KEYWORD_NAME: name, 'date_format': date_format})
+
+
+def _val_error_invalid_time_format(name: str, value: any, time_format: str):
+    time_format = _format_datetime_format(time_format)
+    raise SchemaValidationFailure(lang.VAL_ERROR_PARAM_TIME_FORMAT, {KEYWORD_NAME: name, 'time_format': time_format})
+
+
+def _val_error_invalid_datetime_format(name: str, value: any, datetime_format: str):
+    datetime_format = _format_datetime_format(datetime_format)
+    raise SchemaValidationFailure(lang.VAL_ERROR_PARAM_DATETIME_FORMAT, {KEYWORD_NAME: name, 'datetime_format': datetime_format})
+
+
+def _get_type_str(name: str, value: any, sformat: str or None, join: str or None) -> str or None:
     if value is None:
         return None
     if isinstance(value, int) or isinstance(value, float) or isinstance(value, bool):
-        value = str(value)
+        if sformat is not None:
+            return ("{" + sformat + "}").format(value)
+        return str(value)
+    if isinstance(value, date) or isinstance(value, time) or isinstance(value, datetime):
+        if sformat is not None:
+            if isinstance(value, datetime):
+                return value.strftime(sformat)
+            if isinstance(value, date):
+                return datetime.combine(value, datetime.min.time()).strftime(sformat)
+            return datetime.combine(datetime.today(), value).strftime(sformat)
+        return str(value)
+    if isinstance(value, list):
+        if join is not None:
+            return join.join(value)
+        return str(value)
     if not isinstance(value, str):
         _val_error_unexpected_type(name, value, KEYWORD_TYPE_STR)
+    if sformat is not None:
+        value = ("{" + sformat + "}").format(value)
     return value
 
 
@@ -317,6 +437,54 @@ def _get_type_bool(name: str, value: any) -> bool or None:
     return value
 
 
+def _get_type_date(name: str, value: any, date_format: str) -> date or None:
+    if value is None:
+        return None
+    if isinstance(value, date):
+        return value
+    elif isinstance(value, datetime):
+        return value.date()
+    elif not isinstance(value, str):
+        _val_error_unexpected_type(name, value, KEYWORD_TYPE_DATE)
+    try:
+        value = datetime.strptime(value, date_format)
+    except ValueError:
+        _val_error_invalid_date_format(name, value, date_format)
+    return value.date()
+
+
+def _get_type_time(name: str, value: any, time_format: str) -> time or None:
+    if value is None:
+        return None
+    if isinstance(value, date):
+        return value
+    elif isinstance(value, datetime):
+        return value.time()
+    elif not isinstance(value, str):
+        _val_error_unexpected_type(name, value, KEYWORD_TYPE_TIME)
+    try:
+        value = datetime.strptime(value, time_format)
+    except ValueError:
+        _val_error_invalid_time_format(name, value, time_format)
+    return value.time()
+
+
+def _get_type_datetime(name: str, value: any, datetime_format: str) -> date or None:
+    if value is None:
+        return None
+    if isinstance(value, datetime):
+        return value
+    elif isinstance(value, date):
+        return datetime.combine(value, datetime.min.time())
+    elif not isinstance(value, str):
+        _val_error_unexpected_type(name, value, KEYWORD_TYPE_DATETIME)
+    try:
+        value = datetime.strptime(value, datetime_format)
+    except ValueError:
+        _val_error_invalid_datetime_format(name, value, datetime_format)
+    return value
+
+
 def _get_type_list(name: str, value: any, separator: str) -> list or None:
     if value is None:
         return None
